package at.tuwien.endpoints;

import at.tuwien.SortType;
import at.tuwien.api.database.query.ImportDto;
import at.tuwien.api.database.query.QueryResultDto;
import at.tuwien.api.database.table.TableCsvDeleteDto;
import at.tuwien.api.database.table.TableCsvDto;
import at.tuwien.api.database.table.TableCsvUpdateDto;
import at.tuwien.api.error.ApiErrorDto;
import at.tuwien.entities.database.Database;
import at.tuwien.exception.*;
import at.tuwien.service.DatabaseService;
import at.tuwien.service.QueryService;
import at.tuwien.utils.PrincipalUtil;
import at.tuwien.utils.UserUtil;
import at.tuwien.validation.EndpointValidator;
import io.micrometer.observation.annotation.Observed;
import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.media.Content;
import io.swagger.v3.oas.annotations.media.Schema;
import io.swagger.v3.oas.annotations.responses.ApiResponse;
import io.swagger.v3.oas.annotations.responses.ApiResponses;
import io.swagger.v3.oas.annotations.security.SecurityRequirement;
import jakarta.servlet.ServletRequest;
import jakarta.servlet.http.HttpServletRequest;
import jakarta.validation.Valid;
import jakarta.validation.constraints.NotNull;
import lombok.extern.log4j.Log4j2;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.bind.annotation.*;

import java.security.Principal;
import java.time.Instant;

@Log4j2
@CrossOrigin(origins = "*")
@RestController
@RequestMapping(path = "/api/database/{databaseId}/table/{tableId}/data",
        consumes = MediaType.ALL_VALUE,
        produces = MediaType.APPLICATION_JSON_VALUE)
public class TableDataEndpoint {

    private final QueryService queryService;
    private final DatabaseService databaseService;
    private final EndpointValidator endpointValidator;

    @Autowired
    public TableDataEndpoint(QueryService queryService, DatabaseService databaseService,
                             EndpointValidator endpointValidator) {
        this.queryService = queryService;
        this.databaseService = databaseService;
        this.endpointValidator = endpointValidator;
    }

    @PostMapping
    @Transactional
    @Observed(name = "dbr_table_data_insert")
    @PreAuthorize("hasAuthority('insert-table-data')")
    @Operation(summary = "Insert data", description = "Insert data directly as key-value map tuple",
            security = {@SecurityRequirement(name = "bearerAuth"), @SecurityRequirement(name = "basicAuth")})
    @ApiResponses(value = {
            @ApiResponse(responseCode = "202",
                    description = "Inserted data successfully"),
            @ApiResponse(responseCode = "400",
                    description = "Insert table data is malformed",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = ApiErrorDto.class))}),
            @ApiResponse(responseCode = "403",
                    description = "Access to the database is forbidden",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = ApiErrorDto.class))}),
            @ApiResponse(responseCode = "404",
                    description = "Table or database could not be found",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = ApiErrorDto.class))}),
            @ApiResponse(responseCode = "410",
                    description = "Failed to import LOB-like values",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = ApiErrorDto.class))}),
    })
    public ResponseEntity<Void> insert(@NotNull @PathVariable("databaseId") Long databaseId,
                                    @NotNull @PathVariable("tableId") Long tableId,
                                    @NotNull @Valid @RequestBody TableCsvDto data,
                                    @NotNull Principal principal)
            throws TableNotFoundException, DatabaseNotFoundException, TableMalformedException, NotAllowedException,
            AccessDeniedException, FileStorageException {
        log.debug("endpoint insert data, databaseId={}, tableId={}, data={}, {}", databaseId, tableId, data, PrincipalUtil.formatForDebug(principal));
        /* check */
        endpointValidator.validateOnlyWriteOwnOrWriteAllAccess(databaseId, tableId, principal);
        /* insert */
        queryService.insert(databaseId, tableId, data, principal);
        return ResponseEntity.accepted()
                .build();
    }

    @PutMapping
    @Transactional
    @PreAuthorize("hasAuthority('insert-table-data')")
    @Observed(name = "dbr_table_data_update")
    @Operation(summary = "Update data", security = @SecurityRequirement(name = "bearerAuth"))
    @ApiResponses(value = {
            @ApiResponse(responseCode = "202",
                    description = "Updated data successfully"),
            @ApiResponse(responseCode = "400",
                    description = "Update table data is malformed",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = ApiErrorDto.class))}),
            @ApiResponse(responseCode = "403",
                    description = "Access to the database is forbidden",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = ApiErrorDto.class))}),
            @ApiResponse(responseCode = "404",
                    description = "Table or database could not be found",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = ApiErrorDto.class))}),
            @ApiResponse(responseCode = "410",
                    description = "Failed to import LOB-like values",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = ApiErrorDto.class))}),
    })
    public ResponseEntity<Void> update(@NotNull @PathVariable("databaseId") Long databaseId,
                                       @NotNull @PathVariable("tableId") Long tableId,
                                       @NotNull @Valid @RequestBody TableCsvUpdateDto data,
                                       @NotNull Principal principal)
            throws TableNotFoundException, DatabaseNotFoundException, TableMalformedException,
            ImageNotSupportedException, DatabaseConnectionException, QueryMalformedException,
            UserNotFoundException, NotAllowedException, AccessDeniedException {
        log.debug("endpoint update data, databaseId={}, tableId={}, data={}, {}", databaseId, tableId, data, PrincipalUtil.formatForDebug(principal));
        /* check */
        endpointValidator.validateOnlyWriteOwnOrWriteAllAccess(databaseId, tableId, principal);
        /* update */
        queryService.update(databaseId, tableId, data, principal);
        return ResponseEntity.accepted()
                .build();
    }

    @DeleteMapping
    @Transactional
    @PreAuthorize("hasAuthority('delete-table-data')")
    @Observed(name = "dbr_table_data_delete")
    @Operation(summary = "Delete data", description = "Delete a tuples that match a key-value map",
            security = {@SecurityRequirement(name = "bearerAuth"), @SecurityRequirement(name = "basicAuth")})
    @ApiResponses(value = {
            @ApiResponse(responseCode = "202",
                    description = "Deleted table data successfully"),
            @ApiResponse(responseCode = "400",
                    description = "Table data or query is malformed",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = ApiErrorDto.class))}),
            @ApiResponse(responseCode = "403",
                    description = "Access to the database is forbidden",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = ApiErrorDto.class))}),
            @ApiResponse(responseCode = "404",
                    description = "Table or database could not be found",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = ApiErrorDto.class))}),
    })
    public ResponseEntity<Void> delete(@NotNull @PathVariable("databaseId") Long databaseId,
                                       @NotNull @PathVariable("tableId") Long tableId,
                                       @NotNull @Valid @RequestBody TableCsvDeleteDto data,
                                       @NotNull Principal principal)
            throws TableNotFoundException, DatabaseNotFoundException, TableMalformedException,
            ImageNotSupportedException, QueryMalformedException, NotAllowedException, AccessDeniedException {
        log.debug("endpoint delete data, databaseId={}, tableId={}, data={}, {}", databaseId, tableId, data, PrincipalUtil.formatForDebug(principal));
        /* check */
        endpointValidator.validateOnlyWriteOwnOrWriteAllAccess(databaseId, tableId, principal);
        /* delete */
        queryService.delete(databaseId, tableId, data, principal);
        return ResponseEntity.accepted()
                .build();
    }

    @PostMapping("/import")
    @Transactional
    @PreAuthorize("hasAuthority('insert-table-data')")
    @Observed(name = "dbr_table_data_import")
    @Operation(summary = "Insert data from csv", security = {@SecurityRequirement(name = "bearerAuth"), @SecurityRequirement(name = "basicAuth")})
    @ApiResponses(value = {
            @ApiResponse(responseCode = "202",
                    description = "Import table data successfully"),
            @ApiResponse(responseCode = "400",
                    description = "Table data is malformed",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = ApiErrorDto.class))}),
            @ApiResponse(responseCode = "403",
                    description = "Access to the database is forbidden",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = ApiErrorDto.class))}),
            @ApiResponse(responseCode = "404",
                    description = "Table or database could not be found",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = ApiErrorDto.class))}),
            @ApiResponse(responseCode = "409",
                    description = "Import failed in sidecar",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = ApiErrorDto.class))}),
            @ApiResponse(responseCode = "422",
                    description = "Could not import csv via sidecar",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = ApiErrorDto.class))}),
    })
    public ResponseEntity<Void> importCsv(@NotNull @PathVariable("databaseId") Long databaseId,
                                          @NotNull @PathVariable("tableId") Long tableId,
                                          @NotNull @Valid @RequestBody ImportDto data,
                                          @NotNull Principal principal)
            throws TableNotFoundException, DatabaseNotFoundException, TableMalformedException,
            NotAllowedException, AccessDeniedException, DataDbSidecarException, DataProcessingException {
        log.debug("endpoint insert data from csv, databaseId={}, tableId={}, data={}, {}", databaseId, tableId, data, PrincipalUtil.formatForDebug(principal));
        /* check */
        endpointValidator.validateOnlyWriteOwnOrWriteAllAccess(databaseId, tableId, principal);
        if (data.getNullElement() == null) {
            log.debug("null element not present, default to empty string");
            data.setNullElement("");
        }
        if (data.getLineTermination() == null) {
            log.debug("line termination not present, default to \\r\\n");
            data.setLineTermination("\r\n");
        }
        /* insert */
        queryService.insert(databaseId, tableId, data, principal);
        return ResponseEntity.accepted()
                .build();
    }

    @RequestMapping(method = {RequestMethod.GET, RequestMethod.HEAD})
    @Transactional(readOnly = true)
    @Observed(name = "dbr_table_data_findall")
    @Operation(summary = "Find data", security = {@SecurityRequirement(name = "bearerAuth"), @SecurityRequirement(name = "basicAuth")})
    @ApiResponses(value = {
            @ApiResponse(responseCode = "200",
                    description = "Get table data successfully"),
            @ApiResponse(responseCode = "400",
                    description = "Table data is malformed or image is not supported",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = ApiErrorDto.class))}),
            @ApiResponse(responseCode = "403",
                    description = "Access to the database is forbidden",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = ApiErrorDto.class))}),
            @ApiResponse(responseCode = "404",
                    description = "Table or database could not be found",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = ApiErrorDto.class))}),
            @ApiResponse(responseCode = "409",
                    description = "Result number could not be retrieved from the query store",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = ApiErrorDto.class))}),
    })
    public ResponseEntity<QueryResultDto> getAll(@NotNull @PathVariable("databaseId") Long databaseId,
                                                 @NotNull @PathVariable("tableId") Long tableId,
                                                 @NotNull Principal principal,
                                                 @NotNull HttpServletRequest request,
                                                 @RequestParam(required = false) Instant timestamp,
                                                 @RequestParam(required = false) Long page,
                                                 @RequestParam(required = false) Long size,
                                                 @RequestParam(required = false) SortType sortDirection,
                                                 @RequestParam(required = false) String sortColumn)
            throws TableNotFoundException, DatabaseNotFoundException, ImageNotSupportedException,
            TableMalformedException, PaginationException, QueryMalformedException, SortException, NotAllowedException,
            AccessDeniedException, QueryStoreException {
        log.debug("endpoint find table data, databaseId={}, tableId={}, timestamp={}, page={}, size={}, sortDirection={}, sortColumn={}, {}",
                databaseId, tableId, timestamp, page, size, sortDirection, sortColumn, PrincipalUtil.formatForDebug(principal));
        /* check */
        endpointValidator.validateDataParams(page, size, sortDirection, sortColumn);
        endpointValidator.validateOnlyAccessOrPublic(databaseId, principal);
        final Database database = databaseService.find(databaseId);
        if (!database.getIsPublic() && !UserUtil.hasRole(principal, "view-table-data")) {
            log.error("Failed to view table data: database with id {} is private and user has no authority", databaseId);
            throw new NotAllowedException("Failed to view table data: database with id " + databaseId + " is private and user has no authority");
        }
        /* default */
        if (page == null) {
            log.trace("page is null: default to 0");
            page = 0L;
        }
        if (size == null) {
            log.trace("size is null: default to 10");
            size = 10L;
        }
        /* find */
        final Long count = queryService.tableCount(databaseId, tableId, timestamp, principal);
        final HttpHeaders headers = new HttpHeaders();
        headers.set("X-Count", "" + count);
        if (request.getMethod().equals("GET")) {
            final QueryResultDto response = queryService.tableFindAll(databaseId, tableId, timestamp, page, size, principal);
            log.trace("find table data resulted in result {}", response);
            return ResponseEntity.ok()
                    .headers(headers)
                    .body(response);
        }
        return ResponseEntity.ok()
                .headers(headers)
                .build();
    }

}
